# Autogenerated. Do not modify.
#
# Profile: 21.60
from typing import List as list
from typing import Optional

from fit_tool.base_type import BaseType
from fit_tool.data_message import DataMessage
from fit_tool.definition_message import DefinitionMessage
from fit_tool.developer_field import DeveloperField
from fit_tool.endian import Endian
from fit_tool.field import Field
from fit_tool.profile.profile_type import *


class RecordMessage(DataMessage):
    ID = 20
    NAME = 'record'

    @staticmethod
    def __get_field_size(definition_message: DefinitionMessage, field_id: int) -> int:
        size = 0
        if definition_message:
            field_definition = definition_message.get_field_definition(field_id)
            if field_definition:
                size = field_definition.size

        return size

    def __init__(self, definition_message=None, developer_fields=None, local_id: int = 0,
                 endian: Endian = Endian.LITTLE):
        super().__init__(name=RecordMessage.NAME,
                         global_id=RecordMessage.ID,
                         local_id=definition_message.local_id if definition_message else local_id,
                         endian=definition_message.endian if definition_message else endian,
                         definition_message=definition_message,
                         developer_fields=developer_fields,
                         fields=[
                             TimestampField(
                                 size=self.__get_field_size(definition_message, TimestampField.ID),
                                 growable=definition_message is None),
                             RecordPositionLatField(
                                 size=self.__get_field_size(definition_message, RecordPositionLatField.ID),
                                 growable=definition_message is None),
                             RecordPositionLongField(
                                 size=self.__get_field_size(definition_message, RecordPositionLongField.ID),
                                 growable=definition_message is None),
                             RecordAltitudeField(
                                 size=self.__get_field_size(definition_message, RecordAltitudeField.ID),
                                 growable=definition_message is None),
                             RecordHeartRateField(
                                 size=self.__get_field_size(definition_message, RecordHeartRateField.ID),
                                 growable=definition_message is None),
                             RecordCadenceField(
                                 size=self.__get_field_size(definition_message, RecordCadenceField.ID),
                                 growable=definition_message is None),
                             RecordDistanceField(
                                 size=self.__get_field_size(definition_message, RecordDistanceField.ID),
                                 growable=definition_message is None),
                             RecordSpeedField(
                                 size=self.__get_field_size(definition_message, RecordSpeedField.ID),
                                 growable=definition_message is None),
                             RecordPowerField(
                                 size=self.__get_field_size(definition_message, RecordPowerField.ID),
                                 growable=definition_message is None),
                             RecordCompressedSpeedDistanceField(
                                 size=self.__get_field_size(definition_message, RecordCompressedSpeedDistanceField.ID),
                                 growable=definition_message is None),
                             RecordGradeField(
                                 size=self.__get_field_size(definition_message, RecordGradeField.ID),
                                 growable=definition_message is None),
                             RecordResistanceField(
                                 size=self.__get_field_size(definition_message, RecordResistanceField.ID),
                                 growable=definition_message is None),
                             RecordTimeFromCourseField(
                                 size=self.__get_field_size(definition_message, RecordTimeFromCourseField.ID),
                                 growable=definition_message is None),
                             RecordCycleLengthField(
                                 size=self.__get_field_size(definition_message, RecordCycleLengthField.ID),
                                 growable=definition_message is None),
                             RecordTemperatureField(
                                 size=self.__get_field_size(definition_message, RecordTemperatureField.ID),
                                 growable=definition_message is None),
                             RecordSpeed1sField(
                                 size=self.__get_field_size(definition_message, RecordSpeed1sField.ID),
                                 growable=definition_message is None),
                             RecordCyclesField(
                                 size=self.__get_field_size(definition_message, RecordCyclesField.ID),
                                 growable=definition_message is None),
                             RecordTotalCyclesField(
                                 size=self.__get_field_size(definition_message, RecordTotalCyclesField.ID),
                                 growable=definition_message is None),
                             RecordCompressedAccumulatedPowerField(
                                 size=self.__get_field_size(definition_message,
                                                            RecordCompressedAccumulatedPowerField.ID),
                                 growable=definition_message is None),
                             RecordAccumulatedPowerField(
                                 size=self.__get_field_size(definition_message, RecordAccumulatedPowerField.ID),
                                 growable=definition_message is None),
                             RecordLeftRightBalanceField(
                                 size=self.__get_field_size(definition_message, RecordLeftRightBalanceField.ID),
                                 growable=definition_message is None),
                             RecordGpsAccuracyField(
                                 size=self.__get_field_size(definition_message, RecordGpsAccuracyField.ID),
                                 growable=definition_message is None),
                             RecordVerticalSpeedField(
                                 size=self.__get_field_size(definition_message, RecordVerticalSpeedField.ID),
                                 growable=definition_message is None),
                             RecordCaloriesField(
                                 size=self.__get_field_size(definition_message, RecordCaloriesField.ID),
                                 growable=definition_message is None),
                             RecordVerticalOscillationField(
                                 size=self.__get_field_size(definition_message, RecordVerticalOscillationField.ID),
                                 growable=definition_message is None),
                             RecordStanceTimePercentField(
                                 size=self.__get_field_size(definition_message, RecordStanceTimePercentField.ID),
                                 growable=definition_message is None),
                             RecordStanceTimeField(
                                 size=self.__get_field_size(definition_message, RecordStanceTimeField.ID),
                                 growable=definition_message is None),
                             RecordActivityTypeField(
                                 size=self.__get_field_size(definition_message, RecordActivityTypeField.ID),
                                 growable=definition_message is None),
                             RecordLeftTorqueEffectivenessField(
                                 size=self.__get_field_size(definition_message, RecordLeftTorqueEffectivenessField.ID),
                                 growable=definition_message is None),
                             RecordRightTorqueEffectivenessField(
                                 size=self.__get_field_size(definition_message, RecordRightTorqueEffectivenessField.ID),
                                 growable=definition_message is None),
                             RecordLeftPedalSmoothnessField(
                                 size=self.__get_field_size(definition_message, RecordLeftPedalSmoothnessField.ID),
                                 growable=definition_message is None),
                             RecordRightPedalSmoothnessField(
                                 size=self.__get_field_size(definition_message, RecordRightPedalSmoothnessField.ID),
                                 growable=definition_message is None),
                             RecordCombinedPedalSmoothnessField(
                                 size=self.__get_field_size(definition_message, RecordCombinedPedalSmoothnessField.ID),
                                 growable=definition_message is None),
                             RecordTime128Field(
                                 size=self.__get_field_size(definition_message, RecordTime128Field.ID),
                                 growable=definition_message is None),
                             RecordStrokeTypeField(
                                 size=self.__get_field_size(definition_message, RecordStrokeTypeField.ID),
                                 growable=definition_message is None),
                             RecordZoneField(
                                 size=self.__get_field_size(definition_message, RecordZoneField.ID),
                                 growable=definition_message is None),
                             RecordBallSpeedField(
                                 size=self.__get_field_size(definition_message, RecordBallSpeedField.ID),
                                 growable=definition_message is None),
                             RecordCadence256Field(
                                 size=self.__get_field_size(definition_message, RecordCadence256Field.ID),
                                 growable=definition_message is None),
                             RecordFractionalCadenceField(
                                 size=self.__get_field_size(definition_message, RecordFractionalCadenceField.ID),
                                 growable=definition_message is None),
                             RecordTotalHemoglobinConcField(
                                 size=self.__get_field_size(definition_message, RecordTotalHemoglobinConcField.ID),
                                 growable=definition_message is None),
                             RecordTotalHemoglobinConcMinField(
                                 size=self.__get_field_size(definition_message, RecordTotalHemoglobinConcMinField.ID),
                                 growable=definition_message is None),
                             RecordTotalHemoglobinConcMaxField(
                                 size=self.__get_field_size(definition_message, RecordTotalHemoglobinConcMaxField.ID),
                                 growable=definition_message is None),
                             RecordSaturatedHemoglobinPercentField(
                                 size=self.__get_field_size(definition_message,
                                                            RecordSaturatedHemoglobinPercentField.ID),
                                 growable=definition_message is None),
                             RecordSaturatedHemoglobinPercentMinField(
                                 size=self.__get_field_size(definition_message,
                                                            RecordSaturatedHemoglobinPercentMinField.ID),
                                 growable=definition_message is None),
                             RecordSaturatedHemoglobinPercentMaxField(
                                 size=self.__get_field_size(definition_message,
                                                            RecordSaturatedHemoglobinPercentMaxField.ID),
                                 growable=definition_message is None),
                             RecordDeviceIndexField(
                                 size=self.__get_field_size(definition_message, RecordDeviceIndexField.ID),
                                 growable=definition_message is None),
                             RecordLeftPcoField(
                                 size=self.__get_field_size(definition_message, RecordLeftPcoField.ID),
                                 growable=definition_message is None),
                             RecordRightPcoField(
                                 size=self.__get_field_size(definition_message, RecordRightPcoField.ID),
                                 growable=definition_message is None),
                             RecordLeftPowerPhaseField(
                                 size=self.__get_field_size(definition_message, RecordLeftPowerPhaseField.ID),
                                 growable=definition_message is None),
                             RecordLeftPowerPhasePeakField(
                                 size=self.__get_field_size(definition_message, RecordLeftPowerPhasePeakField.ID),
                                 growable=definition_message is None),
                             RecordRightPowerPhaseField(
                                 size=self.__get_field_size(definition_message, RecordRightPowerPhaseField.ID),
                                 growable=definition_message is None),
                             RecordRightPowerPhasePeakField(
                                 size=self.__get_field_size(definition_message, RecordRightPowerPhasePeakField.ID),
                                 growable=definition_message is None),
                             RecordEnhancedSpeedField(
                                 size=self.__get_field_size(definition_message, RecordEnhancedSpeedField.ID),
                                 growable=definition_message is None),
                             RecordEnhancedAltitudeField(
                                 size=self.__get_field_size(definition_message, RecordEnhancedAltitudeField.ID),
                                 growable=definition_message is None),
                             RecordBatterySocField(
                                 size=self.__get_field_size(definition_message, RecordBatterySocField.ID),
                                 growable=definition_message is None),
                             RecordMotorPowerField(
                                 size=self.__get_field_size(definition_message, RecordMotorPowerField.ID),
                                 growable=definition_message is None),
                             RecordVerticalRatioField(
                                 size=self.__get_field_size(definition_message, RecordVerticalRatioField.ID),
                                 growable=definition_message is None),
                             RecordStanceTimeBalanceField(
                                 size=self.__get_field_size(definition_message, RecordStanceTimeBalanceField.ID),
                                 growable=definition_message is None),
                             RecordStepLengthField(
                                 size=self.__get_field_size(definition_message, RecordStepLengthField.ID),
                                 growable=definition_message is None),
                             RecordAbsolutePressureField(
                                 size=self.__get_field_size(definition_message, RecordAbsolutePressureField.ID),
                                 growable=definition_message is None),
                             RecordDepthField(
                                 size=self.__get_field_size(definition_message, RecordDepthField.ID),
                                 growable=definition_message is None),
                             RecordNextStopDepthField(
                                 size=self.__get_field_size(definition_message, RecordNextStopDepthField.ID),
                                 growable=definition_message is None),
                             RecordNextStopTimeField(
                                 size=self.__get_field_size(definition_message, RecordNextStopTimeField.ID),
                                 growable=definition_message is None),
                             RecordTimeToSurfaceField(
                                 size=self.__get_field_size(definition_message, RecordTimeToSurfaceField.ID),
                                 growable=definition_message is None),
                             RecordNdlTimeField(
                                 size=self.__get_field_size(definition_message, RecordNdlTimeField.ID),
                                 growable=definition_message is None),
                             RecordCnsLoadField(
                                 size=self.__get_field_size(definition_message, RecordCnsLoadField.ID),
                                 growable=definition_message is None),
                             RecordN2LoadField(
                                 size=self.__get_field_size(definition_message, RecordN2LoadField.ID),
                                 growable=definition_message is None),
                             RecordGritField(
                                 size=self.__get_field_size(definition_message, RecordGritField.ID),
                                 growable=definition_message is None),
                             RecordFlowField(
                                 size=self.__get_field_size(definition_message, RecordFlowField.ID),
                                 growable=definition_message is None),
                             RecordEbikeTravelRangeField(
                                 size=self.__get_field_size(definition_message, RecordEbikeTravelRangeField.ID),
                                 growable=definition_message is None),
                             RecordEbikeBatteryLevelField(
                                 size=self.__get_field_size(definition_message, RecordEbikeBatteryLevelField.ID),
                                 growable=definition_message is None),
                             RecordEbikeAssistModeField(
                                 size=self.__get_field_size(definition_message, RecordEbikeAssistModeField.ID),
                                 growable=definition_message is None),
                             RecordEbikeAssistLevelPercentField(
                                 size=self.__get_field_size(definition_message, RecordEbikeAssistLevelPercentField.ID),
                                 growable=definition_message is None),
                             RecordCoreTemperatureField(
                                 size=self.__get_field_size(definition_message, RecordCoreTemperatureField.ID),
                                 growable=definition_message is None)
                         ])

        self.growable = self.definition_message is None

    @classmethod
    def from_bytes(cls, definition_message: DefinitionMessage, developer_fields: list[DeveloperField],
                   bytes_buffer: bytes, offset: int = 0):
        message = cls(definition_message=definition_message, developer_fields=developer_fields)
        message.read_from_bytes(bytes_buffer, offset)
        return message

    # timestamp : milliseconds from January 1st, 1970 at 00:00:00 UTC

    @property
    def timestamp(self) -> Optional[int]:
        field = self.get_field(TimestampField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    # timestamp : milliseconds from January 1st, 1970 at 00:00:00 UTC

    @timestamp.setter
    def timestamp(self, value: int):
        field = self.get_field(TimestampField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def position_lat(self) -> Optional[float]:
        field = self.get_field(RecordPositionLatField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @position_lat.setter
    def position_lat(self, value: float):
        field = self.get_field(RecordPositionLatField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def position_long(self) -> Optional[float]:
        field = self.get_field(RecordPositionLongField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @position_long.setter
    def position_long(self, value: float):
        field = self.get_field(RecordPositionLongField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def altitude(self) -> Optional[float]:
        field = self.get_field(RecordAltitudeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @altitude.setter
    def altitude(self, value: float):
        field = self.get_field(RecordAltitudeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def heart_rate(self) -> Optional[int]:
        field = self.get_field(RecordHeartRateField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @heart_rate.setter
    def heart_rate(self, value: int):
        field = self.get_field(RecordHeartRateField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def cadence(self) -> Optional[int]:
        field = self.get_field(RecordCadenceField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @cadence.setter
    def cadence(self, value: int):
        field = self.get_field(RecordCadenceField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def distance(self) -> Optional[float]:
        field = self.get_field(RecordDistanceField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @distance.setter
    def distance(self, value: float):
        field = self.get_field(RecordDistanceField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def speed(self) -> Optional[float]:
        field = self.get_field(RecordSpeedField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @speed.setter
    def speed(self, value: float):
        field = self.get_field(RecordSpeedField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def power(self) -> Optional[int]:
        field = self.get_field(RecordPowerField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @power.setter
    def power(self, value: int):
        field = self.get_field(RecordPowerField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def compressed_speed_distance(self) -> Optional[bytes]:
        field = self.get_field(RecordCompressedSpeedDistanceField.ID)
        if field and field.is_valid():
            return field.get_values()
        else:
            return None

    @compressed_speed_distance.setter
    def compressed_speed_distance(self, value: bytes):
        field = self.get_field(RecordCompressedSpeedDistanceField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                field.set_values(value)

    @property
    def grade(self) -> Optional[float]:
        field = self.get_field(RecordGradeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @grade.setter
    def grade(self, value: float):
        field = self.get_field(RecordGradeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def resistance(self) -> Optional[int]:
        field = self.get_field(RecordResistanceField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @resistance.setter
    def resistance(self, value: int):
        field = self.get_field(RecordResistanceField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def time_from_course(self) -> Optional[float]:
        field = self.get_field(RecordTimeFromCourseField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @time_from_course.setter
    def time_from_course(self, value: float):
        field = self.get_field(RecordTimeFromCourseField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def cycle_length(self) -> Optional[float]:
        field = self.get_field(RecordCycleLengthField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @cycle_length.setter
    def cycle_length(self, value: float):
        field = self.get_field(RecordCycleLengthField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def temperature(self) -> Optional[int]:
        field = self.get_field(RecordTemperatureField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @temperature.setter
    def temperature(self, value: int):
        field = self.get_field(RecordTemperatureField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def speed_1s(self) -> Optional[list[float]]:
        field = self.get_field(RecordSpeed1sField.ID)
        if field and field.is_valid():
            return field.get_values()
        else:
            return None

    @speed_1s.setter
    def speed_1s(self, value: list[float]):
        field = self.get_field(RecordSpeed1sField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                field.set_values(value)

    @property
    def cycles(self) -> Optional[int]:
        field = self.get_field(RecordCyclesField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @cycles.setter
    def cycles(self, value: int):
        field = self.get_field(RecordCyclesField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def total_cycles(self) -> Optional[int]:
        field = self.get_field(RecordTotalCyclesField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @total_cycles.setter
    def total_cycles(self, value: int):
        field = self.get_field(RecordTotalCyclesField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def compressed_accumulated_power(self) -> Optional[int]:
        field = self.get_field(RecordCompressedAccumulatedPowerField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @compressed_accumulated_power.setter
    def compressed_accumulated_power(self, value: int):
        field = self.get_field(RecordCompressedAccumulatedPowerField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def accumulated_power(self) -> Optional[int]:
        field = self.get_field(RecordAccumulatedPowerField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @accumulated_power.setter
    def accumulated_power(self, value: int):
        field = self.get_field(RecordAccumulatedPowerField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def left_right_balance(self) -> Optional[int]:
        field = self.get_field(RecordLeftRightBalanceField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @left_right_balance.setter
    def left_right_balance(self, value: int):
        field = self.get_field(RecordLeftRightBalanceField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def gps_accuracy(self) -> Optional[int]:
        field = self.get_field(RecordGpsAccuracyField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @gps_accuracy.setter
    def gps_accuracy(self, value: int):
        field = self.get_field(RecordGpsAccuracyField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def vertical_speed(self) -> Optional[float]:
        field = self.get_field(RecordVerticalSpeedField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @vertical_speed.setter
    def vertical_speed(self, value: float):
        field = self.get_field(RecordVerticalSpeedField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def calories(self) -> Optional[int]:
        field = self.get_field(RecordCaloriesField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @calories.setter
    def calories(self, value: int):
        field = self.get_field(RecordCaloriesField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def vertical_oscillation(self) -> Optional[float]:
        field = self.get_field(RecordVerticalOscillationField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @vertical_oscillation.setter
    def vertical_oscillation(self, value: float):
        field = self.get_field(RecordVerticalOscillationField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def stance_time_percent(self) -> Optional[float]:
        field = self.get_field(RecordStanceTimePercentField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @stance_time_percent.setter
    def stance_time_percent(self, value: float):
        field = self.get_field(RecordStanceTimePercentField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def stance_time(self) -> Optional[float]:
        field = self.get_field(RecordStanceTimeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @stance_time.setter
    def stance_time(self, value: float):
        field = self.get_field(RecordStanceTimeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def activity_type(self) -> Optional[ActivityType]:
        field = self.get_field(RecordActivityTypeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @activity_type.setter
    def activity_type(self, value: ActivityType):
        field = self.get_field(RecordActivityTypeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def left_torque_effectiveness(self) -> Optional[float]:
        field = self.get_field(RecordLeftTorqueEffectivenessField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @left_torque_effectiveness.setter
    def left_torque_effectiveness(self, value: float):
        field = self.get_field(RecordLeftTorqueEffectivenessField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def right_torque_effectiveness(self) -> Optional[float]:
        field = self.get_field(RecordRightTorqueEffectivenessField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @right_torque_effectiveness.setter
    def right_torque_effectiveness(self, value: float):
        field = self.get_field(RecordRightTorqueEffectivenessField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def left_pedal_smoothness(self) -> Optional[float]:
        field = self.get_field(RecordLeftPedalSmoothnessField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @left_pedal_smoothness.setter
    def left_pedal_smoothness(self, value: float):
        field = self.get_field(RecordLeftPedalSmoothnessField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def right_pedal_smoothness(self) -> Optional[float]:
        field = self.get_field(RecordRightPedalSmoothnessField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @right_pedal_smoothness.setter
    def right_pedal_smoothness(self, value: float):
        field = self.get_field(RecordRightPedalSmoothnessField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def combined_pedal_smoothness(self) -> Optional[float]:
        field = self.get_field(RecordCombinedPedalSmoothnessField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @combined_pedal_smoothness.setter
    def combined_pedal_smoothness(self, value: float):
        field = self.get_field(RecordCombinedPedalSmoothnessField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def time128(self) -> Optional[float]:
        field = self.get_field(RecordTime128Field.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @time128.setter
    def time128(self, value: float):
        field = self.get_field(RecordTime128Field.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def stroke_type(self) -> Optional[StrokeType]:
        field = self.get_field(RecordStrokeTypeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @stroke_type.setter
    def stroke_type(self, value: StrokeType):
        field = self.get_field(RecordStrokeTypeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def zone(self) -> Optional[int]:
        field = self.get_field(RecordZoneField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @zone.setter
    def zone(self, value: int):
        field = self.get_field(RecordZoneField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def ball_speed(self) -> Optional[float]:
        field = self.get_field(RecordBallSpeedField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @ball_speed.setter
    def ball_speed(self, value: float):
        field = self.get_field(RecordBallSpeedField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def cadence256(self) -> Optional[float]:
        field = self.get_field(RecordCadence256Field.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @cadence256.setter
    def cadence256(self, value: float):
        field = self.get_field(RecordCadence256Field.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def fractional_cadence(self) -> Optional[float]:
        field = self.get_field(RecordFractionalCadenceField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @fractional_cadence.setter
    def fractional_cadence(self, value: float):
        field = self.get_field(RecordFractionalCadenceField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def total_hemoglobin_conc(self) -> Optional[float]:
        field = self.get_field(RecordTotalHemoglobinConcField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @total_hemoglobin_conc.setter
    def total_hemoglobin_conc(self, value: float):
        field = self.get_field(RecordTotalHemoglobinConcField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def total_hemoglobin_conc_min(self) -> Optional[float]:
        field = self.get_field(RecordTotalHemoglobinConcMinField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @total_hemoglobin_conc_min.setter
    def total_hemoglobin_conc_min(self, value: float):
        field = self.get_field(RecordTotalHemoglobinConcMinField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def total_hemoglobin_conc_max(self) -> Optional[float]:
        field = self.get_field(RecordTotalHemoglobinConcMaxField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @total_hemoglobin_conc_max.setter
    def total_hemoglobin_conc_max(self, value: float):
        field = self.get_field(RecordTotalHemoglobinConcMaxField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def saturated_hemoglobin_percent(self) -> Optional[float]:
        field = self.get_field(RecordSaturatedHemoglobinPercentField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @saturated_hemoglobin_percent.setter
    def saturated_hemoglobin_percent(self, value: float):
        field = self.get_field(RecordSaturatedHemoglobinPercentField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def saturated_hemoglobin_percent_min(self) -> Optional[float]:
        field = self.get_field(RecordSaturatedHemoglobinPercentMinField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @saturated_hemoglobin_percent_min.setter
    def saturated_hemoglobin_percent_min(self, value: float):
        field = self.get_field(RecordSaturatedHemoglobinPercentMinField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def saturated_hemoglobin_percent_max(self) -> Optional[float]:
        field = self.get_field(RecordSaturatedHemoglobinPercentMaxField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @saturated_hemoglobin_percent_max.setter
    def saturated_hemoglobin_percent_max(self, value: float):
        field = self.get_field(RecordSaturatedHemoglobinPercentMaxField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def device_index(self) -> Optional[int]:
        field = self.get_field(RecordDeviceIndexField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @device_index.setter
    def device_index(self, value: int):
        field = self.get_field(RecordDeviceIndexField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def left_pco(self) -> Optional[int]:
        field = self.get_field(RecordLeftPcoField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @left_pco.setter
    def left_pco(self, value: int):
        field = self.get_field(RecordLeftPcoField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def right_pco(self) -> Optional[int]:
        field = self.get_field(RecordRightPcoField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @right_pco.setter
    def right_pco(self, value: int):
        field = self.get_field(RecordRightPcoField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def left_power_phase(self) -> Optional[list[float]]:
        field = self.get_field(RecordLeftPowerPhaseField.ID)
        if field and field.is_valid():
            return field.get_values()
        else:
            return None

    @left_power_phase.setter
    def left_power_phase(self, value: list[float]):
        field = self.get_field(RecordLeftPowerPhaseField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                field.set_values(value)

    @property
    def left_power_phase_peak(self) -> Optional[list[float]]:
        field = self.get_field(RecordLeftPowerPhasePeakField.ID)
        if field and field.is_valid():
            return field.get_values()
        else:
            return None

    @left_power_phase_peak.setter
    def left_power_phase_peak(self, value: list[float]):
        field = self.get_field(RecordLeftPowerPhasePeakField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                field.set_values(value)

    @property
    def right_power_phase(self) -> Optional[list[float]]:
        field = self.get_field(RecordRightPowerPhaseField.ID)
        if field and field.is_valid():
            return field.get_values()
        else:
            return None

    @right_power_phase.setter
    def right_power_phase(self, value: list[float]):
        field = self.get_field(RecordRightPowerPhaseField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                field.set_values(value)

    @property
    def right_power_phase_peak(self) -> Optional[list[float]]:
        field = self.get_field(RecordRightPowerPhasePeakField.ID)
        if field and field.is_valid():
            return field.get_values()
        else:
            return None

    @right_power_phase_peak.setter
    def right_power_phase_peak(self, value: list[float]):
        field = self.get_field(RecordRightPowerPhasePeakField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                field.set_values(value)

    @property
    def enhanced_speed(self) -> Optional[float]:
        field = self.get_field(RecordEnhancedSpeedField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @enhanced_speed.setter
    def enhanced_speed(self, value: float):
        field = self.get_field(RecordEnhancedSpeedField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def enhanced_altitude(self) -> Optional[float]:
        field = self.get_field(RecordEnhancedAltitudeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @enhanced_altitude.setter
    def enhanced_altitude(self, value: float):
        field = self.get_field(RecordEnhancedAltitudeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def battery_soc(self) -> Optional[float]:
        field = self.get_field(RecordBatterySocField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @battery_soc.setter
    def battery_soc(self, value: float):
        field = self.get_field(RecordBatterySocField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def motor_power(self) -> Optional[int]:
        field = self.get_field(RecordMotorPowerField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @motor_power.setter
    def motor_power(self, value: int):
        field = self.get_field(RecordMotorPowerField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def vertical_ratio(self) -> Optional[float]:
        field = self.get_field(RecordVerticalRatioField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @vertical_ratio.setter
    def vertical_ratio(self, value: float):
        field = self.get_field(RecordVerticalRatioField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def stance_time_balance(self) -> Optional[float]:
        field = self.get_field(RecordStanceTimeBalanceField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @stance_time_balance.setter
    def stance_time_balance(self, value: float):
        field = self.get_field(RecordStanceTimeBalanceField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def step_length(self) -> Optional[float]:
        field = self.get_field(RecordStepLengthField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @step_length.setter
    def step_length(self, value: float):
        field = self.get_field(RecordStepLengthField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def absolute_pressure(self) -> Optional[int]:
        field = self.get_field(RecordAbsolutePressureField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @absolute_pressure.setter
    def absolute_pressure(self, value: int):
        field = self.get_field(RecordAbsolutePressureField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def depth(self) -> Optional[float]:
        field = self.get_field(RecordDepthField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @depth.setter
    def depth(self, value: float):
        field = self.get_field(RecordDepthField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def next_stop_depth(self) -> Optional[float]:
        field = self.get_field(RecordNextStopDepthField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @next_stop_depth.setter
    def next_stop_depth(self, value: float):
        field = self.get_field(RecordNextStopDepthField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def next_stop_time(self) -> Optional[int]:
        field = self.get_field(RecordNextStopTimeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @next_stop_time.setter
    def next_stop_time(self, value: int):
        field = self.get_field(RecordNextStopTimeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def time_to_surface(self) -> Optional[int]:
        field = self.get_field(RecordTimeToSurfaceField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @time_to_surface.setter
    def time_to_surface(self, value: int):
        field = self.get_field(RecordTimeToSurfaceField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def ndl_time(self) -> Optional[int]:
        field = self.get_field(RecordNdlTimeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @ndl_time.setter
    def ndl_time(self, value: int):
        field = self.get_field(RecordNdlTimeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def cns_load(self) -> Optional[int]:
        field = self.get_field(RecordCnsLoadField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @cns_load.setter
    def cns_load(self, value: int):
        field = self.get_field(RecordCnsLoadField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def n2_load(self) -> Optional[int]:
        field = self.get_field(RecordN2LoadField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @n2_load.setter
    def n2_load(self, value: int):
        field = self.get_field(RecordN2LoadField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def grit(self) -> Optional[float]:
        field = self.get_field(RecordGritField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @grit.setter
    def grit(self, value: float):
        field = self.get_field(RecordGritField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def flow(self) -> Optional[float]:
        field = self.get_field(RecordFlowField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @flow.setter
    def flow(self, value: float):
        field = self.get_field(RecordFlowField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def ebike_travel_range(self) -> Optional[int]:
        field = self.get_field(RecordEbikeTravelRangeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @ebike_travel_range.setter
    def ebike_travel_range(self, value: int):
        field = self.get_field(RecordEbikeTravelRangeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def ebike_battery_level(self) -> Optional[int]:
        field = self.get_field(RecordEbikeBatteryLevelField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @ebike_battery_level.setter
    def ebike_battery_level(self, value: int):
        field = self.get_field(RecordEbikeBatteryLevelField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def ebike_assist_mode(self) -> Optional[int]:
        field = self.get_field(RecordEbikeAssistModeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @ebike_assist_mode.setter
    def ebike_assist_mode(self, value: int):
        field = self.get_field(RecordEbikeAssistModeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def ebike_assist_level_percent(self) -> Optional[int]:
        field = self.get_field(RecordEbikeAssistLevelPercentField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @ebike_assist_level_percent.setter
    def ebike_assist_level_percent(self, value: int):
        field = self.get_field(RecordEbikeAssistLevelPercentField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def core_temperature(self) -> Optional[float]:
        field = self.get_field(RecordCoreTemperatureField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @core_temperature.setter
    def core_temperature(self, value: float):
        field = self.get_field(RecordCoreTemperatureField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)


class TimestampField(Field):
    ID = 253

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='timestamp',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=-631065600000,
            scale=0.001,
            size=size,
            units='ms',
            type_name='date_time',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordPositionLatField(Field):
    ID = 0

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='position_lat',
            field_id=self.ID,
            base_type=BaseType.SINT32,
            offset=0,
            scale=11930464.711111112,
            size=size,
            units='degrees',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordPositionLongField(Field):
    ID = 1

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='position_long',
            field_id=self.ID,
            base_type=BaseType.SINT32,
            offset=0,
            scale=11930464.711111112,
            size=size,
            units='degrees',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordAltitudeField(Field):
    ID = 2

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='altitude',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=500,
            scale=5,
            size=size,
            units='m',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordHeartRateField(Field):
    ID = 3

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='heart_rate',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            units='bpm',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordCadenceField(Field):
    ID = 4

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='cadence',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            units='rpm',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordDistanceField(Field):
    ID = 5

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='distance',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=100,
            size=size,
            units='m',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordSpeedField(Field):
    ID = 6

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='speed',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1000,
            size=size,
            units='m/s',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordPowerField(Field):
    ID = 7

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='power',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1,
            size=size,
            units='watts',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordCompressedSpeedDistanceField(Field):
    ID = 8

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='compressed_speed_distance',
            field_id=self.ID,
            base_type=BaseType.BYTE,
            offset=0,
            scale=1,
            size=size,
            units='m/s,m',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordGradeField(Field):
    ID = 9

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='grade',
            field_id=self.ID,
            base_type=BaseType.SINT16,
            offset=0,
            scale=100,
            size=size,
            units='%',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordResistanceField(Field):
    ID = 10

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='resistance',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class RecordTimeFromCourseField(Field):
    ID = 11

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='time_from_course',
            field_id=self.ID,
            base_type=BaseType.SINT32,
            offset=0,
            scale=1000,
            size=size,
            units='s',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordCycleLengthField(Field):
    ID = 12

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='cycle_length',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=100,
            size=size,
            units='m',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordTemperatureField(Field):
    ID = 13

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='temperature',
            field_id=self.ID,
            base_type=BaseType.SINT8,
            offset=0,
            scale=1,
            size=size,
            units='C',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordSpeed1sField(Field):
    ID = 17

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='speed_1s',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=16,
            size=size,
            units='m/s',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordCyclesField(Field):
    ID = 18

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='cycles',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            units='cycles',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordTotalCyclesField(Field):
    ID = 19

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='total_cycles',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=1,
            size=size,
            units='cycles',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordCompressedAccumulatedPowerField(Field):
    ID = 28

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='compressed_accumulated_power',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1,
            size=size,
            units='watts',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordAccumulatedPowerField(Field):
    ID = 29

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='accumulated_power',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=1,
            size=size,
            units='watts',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordLeftRightBalanceField(Field):
    ID = 30

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='left_right_balance',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class RecordGpsAccuracyField(Field):
    ID = 31

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='gps_accuracy',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            units='m',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordVerticalSpeedField(Field):
    ID = 32

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='vertical_speed',
            field_id=self.ID,
            base_type=BaseType.SINT16,
            offset=0,
            scale=1000,
            size=size,
            units='m/s',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordCaloriesField(Field):
    ID = 33

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='calories',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1,
            size=size,
            units='kcal',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordVerticalOscillationField(Field):
    ID = 39

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='vertical_oscillation',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=10,
            size=size,
            units='mm',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordStanceTimePercentField(Field):
    ID = 40

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='stance_time_percent',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=100,
            size=size,
            units='percent',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordStanceTimeField(Field):
    ID = 41

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='stance_time',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=10,
            size=size,
            units='ms',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordActivityTypeField(Field):
    ID = 42

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='activity_type',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class RecordLeftTorqueEffectivenessField(Field):
    ID = 43

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='left_torque_effectiveness',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=2,
            size=size,
            units='percent',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordRightTorqueEffectivenessField(Field):
    ID = 44

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='right_torque_effectiveness',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=2,
            size=size,
            units='percent',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordLeftPedalSmoothnessField(Field):
    ID = 45

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='left_pedal_smoothness',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=2,
            size=size,
            units='percent',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordRightPedalSmoothnessField(Field):
    ID = 46

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='right_pedal_smoothness',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=2,
            size=size,
            units='percent',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordCombinedPedalSmoothnessField(Field):
    ID = 47

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='combined_pedal_smoothness',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=2,
            size=size,
            units='percent',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordTime128Field(Field):
    ID = 48

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='time128',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=128,
            size=size,
            units='s',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordStrokeTypeField(Field):
    ID = 49

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='stroke_type',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class RecordZoneField(Field):
    ID = 50

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='zone',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class RecordBallSpeedField(Field):
    ID = 51

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='ball_speed',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=100,
            size=size,
            units='m/s',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordCadence256Field(Field):
    ID = 52

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='cadence256',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=256,
            size=size,
            units='rpm',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordFractionalCadenceField(Field):
    ID = 53

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='fractional_cadence',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=128,
            size=size,
            units='rpm',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordTotalHemoglobinConcField(Field):
    ID = 54

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='total_hemoglobin_conc',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=100,
            size=size,
            units='g/dL',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordTotalHemoglobinConcMinField(Field):
    ID = 55

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='total_hemoglobin_conc_min',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=100,
            size=size,
            units='g/dL',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordTotalHemoglobinConcMaxField(Field):
    ID = 56

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='total_hemoglobin_conc_max',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=100,
            size=size,
            units='g/dL',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordSaturatedHemoglobinPercentField(Field):
    ID = 57

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='saturated_hemoglobin_percent',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=10,
            size=size,
            units='%',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordSaturatedHemoglobinPercentMinField(Field):
    ID = 58

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='saturated_hemoglobin_percent_min',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=10,
            size=size,
            units='%',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordSaturatedHemoglobinPercentMaxField(Field):
    ID = 59

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='saturated_hemoglobin_percent_max',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=10,
            size=size,
            units='%',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordDeviceIndexField(Field):
    ID = 62

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='device_index',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class RecordLeftPcoField(Field):
    ID = 67

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='left_pco',
            field_id=self.ID,
            base_type=BaseType.SINT8,
            offset=0,
            scale=1,
            size=size,
            units='mm',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordRightPcoField(Field):
    ID = 68

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='right_pco',
            field_id=self.ID,
            base_type=BaseType.SINT8,
            offset=0,
            scale=1,
            size=size,
            units='mm',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordLeftPowerPhaseField(Field):
    ID = 69

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='left_power_phase',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=0.7111111,
            size=size,
            units='degrees',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordLeftPowerPhasePeakField(Field):
    ID = 70

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='left_power_phase_peak',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=0.7111111,
            size=size,
            units='degrees',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordRightPowerPhaseField(Field):
    ID = 71

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='right_power_phase',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=0.7111111,
            size=size,
            units='degrees',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordRightPowerPhasePeakField(Field):
    ID = 72

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='right_power_phase_peak',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=0.7111111,
            size=size,
            units='degrees',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordEnhancedSpeedField(Field):
    ID = 73

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='enhanced_speed',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=1000,
            size=size,
            units='m/s',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordEnhancedAltitudeField(Field):
    ID = 78

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='enhanced_altitude',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=500,
            scale=5,
            size=size,
            units='m',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordBatterySocField(Field):
    ID = 81

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='battery_soc',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=2,
            size=size,
            units='percent',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordMotorPowerField(Field):
    ID = 82

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='motor_power',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1,
            size=size,
            units='watts',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordVerticalRatioField(Field):
    ID = 83

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='vertical_ratio',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=100,
            size=size,
            units='percent',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordStanceTimeBalanceField(Field):
    ID = 84

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='stance_time_balance',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=100,
            size=size,
            units='percent',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordStepLengthField(Field):
    ID = 85

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='step_length',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=10,
            size=size,
            units='mm',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordAbsolutePressureField(Field):
    ID = 91

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='absolute_pressure',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=1,
            size=size,
            units='Pa',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordDepthField(Field):
    ID = 92

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='depth',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=1000,
            size=size,
            units='m',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordNextStopDepthField(Field):
    ID = 93

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='next_stop_depth',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=1000,
            size=size,
            units='m',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordNextStopTimeField(Field):
    ID = 94

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='next_stop_time',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=1,
            size=size,
            units='s',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordTimeToSurfaceField(Field):
    ID = 95

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='time_to_surface',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=1,
            size=size,
            units='s',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordNdlTimeField(Field):
    ID = 96

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='ndl_time',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=1,
            size=size,
            units='s',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordCnsLoadField(Field):
    ID = 97

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='cns_load',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            units='percent',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordN2LoadField(Field):
    ID = 98

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='n2_load',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1,
            size=size,
            units='percent',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordGritField(Field):
    ID = 114

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='grit',
            field_id=self.ID,
            base_type=BaseType.FLOAT32,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class RecordFlowField(Field):
    ID = 115

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='flow',
            field_id=self.ID,
            base_type=BaseType.FLOAT32,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class RecordEbikeTravelRangeField(Field):
    ID = 117

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='ebike_travel_range',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1,
            size=size,
            units='km',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordEbikeBatteryLevelField(Field):
    ID = 118

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='ebike_battery_level',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            units='percent',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordEbikeAssistModeField(Field):
    ID = 119

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='ebike_assist_mode',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            units='depends on sensor',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordEbikeAssistLevelPercentField(Field):
    ID = 120

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='ebike_assist_level_percent',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            units='percent',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )


class RecordCoreTemperatureField(Field):
    ID = 139

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='core_temperature',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=100,
            size=size,
            units='C',
            type_name='',
            growable=growable,
            sub_fields=[
            ]
        )



# Autogenerated. Do not modify.
#
# Profile: 21.60
from typing import List as list
from typing import Optional

from fit_tool.base_type import BaseType
from fit_tool.data_message import DataMessage
from fit_tool.definition_message import DefinitionMessage
from fit_tool.developer_field import DeveloperField
from fit_tool.endian import Endian
from fit_tool.field import Field


class ConnectivityMessage(DataMessage):
    ID = 127
    NAME = 'connectivity'

    @staticmethod
    def __get_field_size(definition_message: DefinitionMessage, field_id: int) -> int:
        size = 0
        if definition_message:
            field_definition = definition_message.get_field_definition(field_id)
            if field_definition:
                size = field_definition.size

        return size

    def __init__(self, definition_message=None, developer_fields=None, local_id: int = 0,
                 endian: Endian = Endian.LITTLE):
        super().__init__(name=ConnectivityMessage.NAME,
                         global_id=ConnectivityMessage.ID,
                         local_id=definition_message.local_id if definition_message else local_id,
                         endian=definition_message.endian if definition_message else endian,
                         definition_message=definition_message,
                         developer_fields=developer_fields,
                         fields=[
                             ConnectivityBluetoothEnabledField(
                                 size=self.__get_field_size(definition_message, ConnectivityBluetoothEnabledField.ID),
                                 growable=definition_message is None),
                             ConnectivityBluetoothLeEnabledField(
                                 size=self.__get_field_size(definition_message, ConnectivityBluetoothLeEnabledField.ID),
                                 growable=definition_message is None),
                             ConnectivityAntEnabledField(
                                 size=self.__get_field_size(definition_message, ConnectivityAntEnabledField.ID),
                                 growable=definition_message is None),
                             ConnectivityNameField(
                                 size=self.__get_field_size(definition_message, ConnectivityNameField.ID),
                                 growable=definition_message is None),
                             ConnectivityLiveTrackingEnabledField(
                                 size=self.__get_field_size(definition_message,
                                                            ConnectivityLiveTrackingEnabledField.ID),
                                 growable=definition_message is None),
                             ConnectivityWeatherConditionsEnabledField(
                                 size=self.__get_field_size(definition_message,
                                                            ConnectivityWeatherConditionsEnabledField.ID),
                                 growable=definition_message is None),
                             ConnectivityWeatherAlertsEnabledField(
                                 size=self.__get_field_size(definition_message,
                                                            ConnectivityWeatherAlertsEnabledField.ID),
                                 growable=definition_message is None),
                             ConnectivityAutoActivityUploadEnabledField(
                                 size=self.__get_field_size(definition_message,
                                                            ConnectivityAutoActivityUploadEnabledField.ID),
                                 growable=definition_message is None),
                             ConnectivityCourseDownloadEnabledField(
                                 size=self.__get_field_size(definition_message,
                                                            ConnectivityCourseDownloadEnabledField.ID),
                                 growable=definition_message is None),
                             ConnectivityWorkoutDownloadEnabledField(
                                 size=self.__get_field_size(definition_message,
                                                            ConnectivityWorkoutDownloadEnabledField.ID),
                                 growable=definition_message is None),
                             ConnectivityGpsEphemerisDownloadEnabledField(
                                 size=self.__get_field_size(definition_message,
                                                            ConnectivityGpsEphemerisDownloadEnabledField.ID),
                                 growable=definition_message is None),
                             ConnectivityIncidentDetectionEnabledField(
                                 size=self.__get_field_size(definition_message,
                                                            ConnectivityIncidentDetectionEnabledField.ID),
                                 growable=definition_message is None),
                             ConnectivityGrouptrackEnabledField(
                                 size=self.__get_field_size(definition_message, ConnectivityGrouptrackEnabledField.ID),
                                 growable=definition_message is None)
                         ])

        self.growable = self.definition_message is None

    @classmethod
    def from_bytes(cls, definition_message: DefinitionMessage, developer_fields: list[DeveloperField],
                   bytes_buffer: bytes, offset: int = 0):
        message = cls(definition_message=definition_message, developer_fields=developer_fields)
        message.read_from_bytes(bytes_buffer, offset)
        return message

    @property
    def bluetooth_enabled(self) -> Optional[bool]:
        field = self.get_field(ConnectivityBluetoothEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @bluetooth_enabled.setter
    def bluetooth_enabled(self, value: bool):
        field = self.get_field(ConnectivityBluetoothEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def bluetooth_le_enabled(self) -> Optional[bool]:
        field = self.get_field(ConnectivityBluetoothLeEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @bluetooth_le_enabled.setter
    def bluetooth_le_enabled(self, value: bool):
        field = self.get_field(ConnectivityBluetoothLeEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def ant_enabled(self) -> Optional[bool]:
        field = self.get_field(ConnectivityAntEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @ant_enabled.setter
    def ant_enabled(self, value: bool):
        field = self.get_field(ConnectivityAntEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def connectivity_name(self) -> Optional[str]:
        field = self.get_field(ConnectivityNameField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @connectivity_name.setter
    def connectivity_name(self, value: str):
        field = self.get_field(ConnectivityNameField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def live_tracking_enabled(self) -> Optional[bool]:
        field = self.get_field(ConnectivityLiveTrackingEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @live_tracking_enabled.setter
    def live_tracking_enabled(self, value: bool):
        field = self.get_field(ConnectivityLiveTrackingEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def weather_conditions_enabled(self) -> Optional[bool]:
        field = self.get_field(ConnectivityWeatherConditionsEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @weather_conditions_enabled.setter
    def weather_conditions_enabled(self, value: bool):
        field = self.get_field(ConnectivityWeatherConditionsEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def weather_alerts_enabled(self) -> Optional[bool]:
        field = self.get_field(ConnectivityWeatherAlertsEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @weather_alerts_enabled.setter
    def weather_alerts_enabled(self, value: bool):
        field = self.get_field(ConnectivityWeatherAlertsEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def auto_activity_upload_enabled(self) -> Optional[bool]:
        field = self.get_field(ConnectivityAutoActivityUploadEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @auto_activity_upload_enabled.setter
    def auto_activity_upload_enabled(self, value: bool):
        field = self.get_field(ConnectivityAutoActivityUploadEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def course_download_enabled(self) -> Optional[bool]:
        field = self.get_field(ConnectivityCourseDownloadEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @course_download_enabled.setter
    def course_download_enabled(self, value: bool):
        field = self.get_field(ConnectivityCourseDownloadEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def workout_download_enabled(self) -> Optional[bool]:
        field = self.get_field(ConnectivityWorkoutDownloadEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @workout_download_enabled.setter
    def workout_download_enabled(self, value: bool):
        field = self.get_field(ConnectivityWorkoutDownloadEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def gps_ephemeris_download_enabled(self) -> Optional[bool]:
        field = self.get_field(ConnectivityGpsEphemerisDownloadEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @gps_ephemeris_download_enabled.setter
    def gps_ephemeris_download_enabled(self, value: bool):
        field = self.get_field(ConnectivityGpsEphemerisDownloadEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def incident_detection_enabled(self) -> Optional[bool]:
        field = self.get_field(ConnectivityIncidentDetectionEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @incident_detection_enabled.setter
    def incident_detection_enabled(self, value: bool):
        field = self.get_field(ConnectivityIncidentDetectionEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def grouptrack_enabled(self) -> Optional[bool]:
        field = self.get_field(ConnectivityGrouptrackEnabledField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @grouptrack_enabled.setter
    def grouptrack_enabled(self, value: bool):
        field = self.get_field(ConnectivityGrouptrackEnabledField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)


class ConnectivityBluetoothEnabledField(Field):
    ID = 0

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='bluetooth_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ConnectivityBluetoothLeEnabledField(Field):
    ID = 1

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='bluetooth_le_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ConnectivityAntEnabledField(Field):
    ID = 2

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='ant_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ConnectivityNameField(Field):
    ID = 3

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='name',
            field_id=self.ID,
            base_type=BaseType.STRING,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ConnectivityLiveTrackingEnabledField(Field):
    ID = 4

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='live_tracking_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ConnectivityWeatherConditionsEnabledField(Field):
    ID = 5

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='weather_conditions_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ConnectivityWeatherAlertsEnabledField(Field):
    ID = 6

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='weather_alerts_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ConnectivityAutoActivityUploadEnabledField(Field):
    ID = 7

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='auto_activity_upload_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ConnectivityCourseDownloadEnabledField(Field):
    ID = 8

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='course_download_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ConnectivityWorkoutDownloadEnabledField(Field):
    ID = 9

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='workout_download_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ConnectivityGpsEphemerisDownloadEnabledField(Field):
    ID = 10

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='gps_ephemeris_download_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ConnectivityIncidentDetectionEnabledField(Field):
    ID = 11

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='incident_detection_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class ConnectivityGrouptrackEnabledField(Field):
    ID = 12

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='grouptrack_enabled',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )



from fit_tool.endian import Endian


class Message:

    def __init__(self, local_id: int = 0, global_id: int = 0, size: int = 0, endian: Endian = Endian.LITTLE):
        # ID that provides an association between the definition message,
        # data message and the FIT message in the Global FIT Profile.
        self.local_id = local_id

        # ID of the FIT message in the Global FIT Profile.
        self.global_id = global_id
        self.endian = endian
        self.size = size

    def to_bytes(self) -> bytes:
        raise Exception('Not implemented')

    def to_row(self) -> list:
        raise Exception('Not implemented')



# Autogenerated. Do not modify.
#
# Profile: 21.60
from typing import List as list
from typing import Optional

from fit_tool.base_type import BaseType
from fit_tool.data_message import DataMessage
from fit_tool.definition_message import DefinitionMessage
from fit_tool.developer_field import DeveloperField
from fit_tool.endian import Endian
from fit_tool.field import Field
from fit_tool.profile.profile_type import *
from fit_tool.sub_field import SubField


class EventMessage(DataMessage):
    ID = 21
    NAME = 'event'

    @staticmethod
    def __get_field_size(definition_message: DefinitionMessage, field_id: int) -> int:
        size = 0
        if definition_message:
            field_definition = definition_message.get_field_definition(field_id)
            if field_definition:
                size = field_definition.size

        return size

    def __init__(self, definition_message=None, developer_fields=None, local_id: int = 0,
                 endian: Endian = Endian.LITTLE):
        super().__init__(name=EventMessage.NAME,
                         global_id=EventMessage.ID,
                         local_id=definition_message.local_id if definition_message else local_id,
                         endian=definition_message.endian if definition_message else endian,
                         definition_message=definition_message,
                         developer_fields=developer_fields,
                         fields=[
                             TimestampField(
                                 size=self.__get_field_size(definition_message, TimestampField.ID),
                                 growable=definition_message is None),
                             EventEventField(
                                 size=self.__get_field_size(definition_message, EventEventField.ID),
                                 growable=definition_message is None),
                             EventEventTypeField(
                                 size=self.__get_field_size(definition_message, EventEventTypeField.ID),
                                 growable=definition_message is None),
                             EventData16Field(
                                 size=self.__get_field_size(definition_message, EventData16Field.ID),
                                 growable=definition_message is None),
                             EventDataField(
                                 size=self.__get_field_size(definition_message, EventDataField.ID),
                                 growable=definition_message is None),
                             EventEventGroupField(
                                 size=self.__get_field_size(definition_message, EventEventGroupField.ID),
                                 growable=definition_message is None),
                             EventScoreField(
                                 size=self.__get_field_size(definition_message, EventScoreField.ID),
                                 growable=definition_message is None),
                             EventOpponentScoreField(
                                 size=self.__get_field_size(definition_message, EventOpponentScoreField.ID),
                                 growable=definition_message is None),
                             EventFrontGearNumField(
                                 size=self.__get_field_size(definition_message, EventFrontGearNumField.ID),
                                 growable=definition_message is None),
                             EventFrontGearField(
                                 size=self.__get_field_size(definition_message, EventFrontGearField.ID),
                                 growable=definition_message is None),
                             EventRearGearNumField(
                                 size=self.__get_field_size(definition_message, EventRearGearNumField.ID),
                                 growable=definition_message is None),
                             EventRearGearField(
                                 size=self.__get_field_size(definition_message, EventRearGearField.ID),
                                 growable=definition_message is None),
                             EventDeviceIndexField(
                                 size=self.__get_field_size(definition_message, EventDeviceIndexField.ID),
                                 growable=definition_message is None),
                             EventRadarThreatLevelMaxField(
                                 size=self.__get_field_size(definition_message, EventRadarThreatLevelMaxField.ID),
                                 growable=definition_message is None),
                             EventRadarThreatCountField(
                                 size=self.__get_field_size(definition_message, EventRadarThreatCountField.ID),
                                 growable=definition_message is None)
                         ])

        self.growable = self.definition_message is None

    @classmethod
    def from_bytes(cls, definition_message: DefinitionMessage, developer_fields: list[DeveloperField],
                   bytes_buffer: bytes, offset: int = 0):
        message = cls(definition_message=definition_message, developer_fields=developer_fields)
        message.read_from_bytes(bytes_buffer, offset)
        return message

    # timestamp : milliseconds from January 1st, 1970 at 00:00:00 UTC

    @property
    def timestamp(self) -> Optional[int]:
        field = self.get_field(TimestampField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    # timestamp : milliseconds from January 1st, 1970 at 00:00:00 UTC

    @timestamp.setter
    def timestamp(self, value: int):
        field = self.get_field(TimestampField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def event(self) -> Optional[Event]:
        field = self.get_field(EventEventField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @event.setter
    def event(self, value: Event):
        field = self.get_field(EventEventField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def event_type(self) -> Optional[EventType]:
        field = self.get_field(EventEventTypeField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @event_type.setter
    def event_type(self, value: EventType):
        field = self.get_field(EventEventTypeField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def data16(self) -> Optional[int]:
        field = self.get_field(EventData16Field.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @data16.setter
    def data16(self, value: int):
        field = self.get_field(EventData16Field.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def data(self) -> Optional[int]:
        field = self.get_field(EventDataField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @data.setter
    def data(self, value: int):
        field = self.get_field(EventDataField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def timer_trigger(self) -> Optional[TimerTrigger]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [0]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @timer_trigger.setter
    def timer_trigger(self, value: TimerTrigger):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def course_point_index(self) -> Optional[int]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [10]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @course_point_index.setter
    def course_point_index(self, value: int):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def battery_level(self) -> Optional[float]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [11]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @battery_level.setter
    def battery_level(self, value: float):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def virtual_partner_speed(self) -> Optional[float]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [12]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @virtual_partner_speed.setter
    def virtual_partner_speed(self, value: float):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def hr_high_alert(self) -> Optional[int]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [13]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @hr_high_alert.setter
    def hr_high_alert(self, value: int):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def hr_low_alert(self) -> Optional[int]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [14]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @hr_low_alert.setter
    def hr_low_alert(self, value: int):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def speed_high_alert(self) -> Optional[float]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [15]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @speed_high_alert.setter
    def speed_high_alert(self, value: float):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def speed_low_alert(self) -> Optional[float]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [16]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @speed_low_alert.setter
    def speed_low_alert(self, value: float):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def cad_high_alert(self) -> Optional[int]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [17]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @cad_high_alert.setter
    def cad_high_alert(self, value: int):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def cad_low_alert(self) -> Optional[int]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [18]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @cad_low_alert.setter
    def cad_low_alert(self, value: int):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def power_high_alert(self) -> Optional[int]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [19]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @power_high_alert.setter
    def power_high_alert(self, value: int):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def power_low_alert(self) -> Optional[int]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [20]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @power_low_alert.setter
    def power_low_alert(self, value: int):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def time_duration_alert(self) -> Optional[float]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [23]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @time_duration_alert.setter
    def time_duration_alert(self, value: float):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def distance_duration_alert(self) -> Optional[float]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [24]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @distance_duration_alert.setter
    def distance_duration_alert(self, value: float):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def calorie_duration_alert(self) -> Optional[int]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [25]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @calorie_duration_alert.setter
    def calorie_duration_alert(self, value: int):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def fitness_equipment_state(self) -> Optional[FitnessEquipmentState]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [27]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @fitness_equipment_state.setter
    def fitness_equipment_state(self, value: FitnessEquipmentState):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def sport_point(self) -> Optional[int]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [33]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @sport_point.setter
    def sport_point(self, value: int):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def gear_change_data(self) -> Optional[int]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [42, 43]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @gear_change_data.setter
    def gear_change_data(self, value: int):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def rider_position(self) -> Optional[RiderPositionType]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [44]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @rider_position.setter
    def rider_position(self, value: RiderPositionType):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def comm_timeout(self) -> Optional[int]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [47]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @comm_timeout.setter
    def comm_timeout(self, value: int):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def radar_threat_alert(self) -> Optional[int]:
        field = self.get_field(EventDataField.ID)
        type_field = self.get_field(EventEventField.ID)

        is_sub_field_valid = type_field and type_field.get_value() in [75]
        if field and field.is_valid() and is_sub_field_valid:
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @radar_threat_alert.setter
    def radar_threat_alert(self, value: int):
        field = self.get_field(EventDataField.ID)
        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def event_group(self) -> Optional[int]:
        field = self.get_field(EventEventGroupField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @event_group.setter
    def event_group(self, value: int):
        field = self.get_field(EventEventGroupField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def score(self) -> Optional[int]:
        field = self.get_field(EventScoreField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @score.setter
    def score(self, value: int):
        field = self.get_field(EventScoreField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def opponent_score(self) -> Optional[int]:
        field = self.get_field(EventOpponentScoreField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @opponent_score.setter
    def opponent_score(self, value: int):
        field = self.get_field(EventOpponentScoreField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def front_gear_num(self) -> Optional[int]:
        field = self.get_field(EventFrontGearNumField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @front_gear_num.setter
    def front_gear_num(self, value: int):
        field = self.get_field(EventFrontGearNumField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def front_gear(self) -> Optional[int]:
        field = self.get_field(EventFrontGearField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @front_gear.setter
    def front_gear(self, value: int):
        field = self.get_field(EventFrontGearField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def rear_gear_num(self) -> Optional[int]:
        field = self.get_field(EventRearGearNumField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @rear_gear_num.setter
    def rear_gear_num(self, value: int):
        field = self.get_field(EventRearGearNumField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def rear_gear(self) -> Optional[int]:
        field = self.get_field(EventRearGearField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @rear_gear.setter
    def rear_gear(self, value: int):
        field = self.get_field(EventRearGearField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def device_index(self) -> Optional[int]:
        field = self.get_field(EventDeviceIndexField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @device_index.setter
    def device_index(self, value: int):
        field = self.get_field(EventDeviceIndexField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def radar_threat_level_max(self) -> Optional[RadarThreatLevelType]:
        field = self.get_field(EventRadarThreatLevelMaxField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @radar_threat_level_max.setter
    def radar_threat_level_max(self, value: RadarThreatLevelType):
        field = self.get_field(EventRadarThreatLevelMaxField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)

    @property
    def radar_threat_count(self) -> Optional[int]:
        field = self.get_field(EventRadarThreatCountField.ID)
        if field and field.is_valid():
            sub_field = field.get_valid_sub_field(self.fields)
            return field.get_value(sub_field=sub_field)
        else:
            return None

    @radar_threat_count.setter
    def radar_threat_count(self, value: int):
        field = self.get_field(EventRadarThreatCountField.ID)

        if field:
            if value is None:
                field.clear()
            else:
                sub_field = field.get_valid_sub_field(self.fields)
                field.set_value(0, value, sub_field)


class TimestampField(Field):
    ID = 253

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='timestamp',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=-631065600000,
            scale=0.001,
            size=size,
            units='ms',
            type_name='date_time',
            growable=growable,
            sub_fields=[
            ]
        )


class EventEventField(Field):
    ID = 0

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='event',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class EventEventTypeField(Field):
    ID = 1

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='event_type',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class EventData16Field(Field):
    ID = 2

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='data16',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class EventDataField(Field):
    ID = 3

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='data',
            field_id=self.ID,
            base_type=BaseType.UINT32,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
                SubField(
                    name='timer_trigger',
                    base_type=BaseType.ENUM,
                    scale=1,
                    offset=0,
                    reference_map={
                        EventEventField.ID: [0]
                    }),
                SubField(
                    name='course_point_index',
                    base_type=BaseType.UINT16,
                    scale=1,
                    offset=0,
                    reference_map={
                        EventEventField.ID: [10]
                    }),
                SubField(
                    name='battery_level',
                    base_type=BaseType.UINT16,
                    scale=1000,
                    offset=0,
                    units='V',
                    reference_map={
                        EventEventField.ID: [11]
                    }),
                SubField(
                    name='virtual_partner_speed',
                    base_type=BaseType.UINT16,
                    scale=1000,
                    offset=0,
                    units='m/s',
                    reference_map={
                        EventEventField.ID: [12]
                    }),
                SubField(
                    name='hr_high_alert',
                    base_type=BaseType.UINT8,
                    scale=1,
                    offset=0,
                    units='bpm',
                    reference_map={
                        EventEventField.ID: [13]
                    }),
                SubField(
                    name='hr_low_alert',
                    base_type=BaseType.UINT8,
                    scale=1,
                    offset=0,
                    units='bpm',
                    reference_map={
                        EventEventField.ID: [14]
                    }),
                SubField(
                    name='speed_high_alert',
                    base_type=BaseType.UINT32,
                    scale=1000,
                    offset=0,
                    units='m/s',
                    reference_map={
                        EventEventField.ID: [15]
                    }),
                SubField(
                    name='speed_low_alert',
                    base_type=BaseType.UINT32,
                    scale=1000,
                    offset=0,
                    units='m/s',
                    reference_map={
                        EventEventField.ID: [16]
                    }),
                SubField(
                    name='cad_high_alert',
                    base_type=BaseType.UINT16,
                    scale=1,
                    offset=0,
                    units='rpm',
                    reference_map={
                        EventEventField.ID: [17]
                    }),
                SubField(
                    name='cad_low_alert',
                    base_type=BaseType.UINT16,
                    scale=1,
                    offset=0,
                    units='rpm',
                    reference_map={
                        EventEventField.ID: [18]
                    }),
                SubField(
                    name='power_high_alert',
                    base_type=BaseType.UINT16,
                    scale=1,
                    offset=0,
                    units='watts',
                    reference_map={
                        EventEventField.ID: [19]
                    }),
                SubField(
                    name='power_low_alert',
                    base_type=BaseType.UINT16,
                    scale=1,
                    offset=0,
                    units='watts',
                    reference_map={
                        EventEventField.ID: [20]
                    }),
                SubField(
                    name='time_duration_alert',
                    base_type=BaseType.UINT32,
                    scale=1000,
                    offset=0,
                    units='s',
                    reference_map={
                        EventEventField.ID: [23]
                    }),
                SubField(
                    name='distance_duration_alert',
                    base_type=BaseType.UINT32,
                    scale=100,
                    offset=0,
                    units='m',
                    reference_map={
                        EventEventField.ID: [24]
                    }),
                SubField(
                    name='calorie_duration_alert',
                    base_type=BaseType.UINT32,
                    scale=1,
                    offset=0,
                    units='calories',
                    reference_map={
                        EventEventField.ID: [25]
                    }),
                SubField(
                    name='fitness_equipment_state',
                    base_type=BaseType.ENUM,
                    scale=1,
                    offset=0,
                    reference_map={
                        EventEventField.ID: [27]
                    }),
                SubField(
                    name='sport_point',
                    base_type=BaseType.UINT32,
                    scale=1,
                    offset=0,
                    reference_map={
                        EventEventField.ID: [33]
                    }),
                SubField(
                    name='gear_change_data',
                    base_type=BaseType.UINT32,
                    scale=1,
                    offset=0,
                    reference_map={
                        EventEventField.ID: [42, 43]
                    }),
                SubField(
                    name='rider_position',
                    base_type=BaseType.ENUM,
                    scale=1,
                    offset=0,
                    reference_map={
                        EventEventField.ID: [44]
                    }),
                SubField(
                    name='comm_timeout',
                    base_type=BaseType.UINT16,
                    scale=1,
                    offset=0,
                    reference_map={
                        EventEventField.ID: [47]
                    }),
                SubField(
                    name='radar_threat_alert',
                    base_type=BaseType.UINT32,
                    scale=1,
                    offset=0,
                    reference_map={
                        EventEventField.ID: [75]
                    })
            ]
        )


class EventEventGroupField(Field):
    ID = 4

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='event_group',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class EventScoreField(Field):
    ID = 7

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='score',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class EventOpponentScoreField(Field):
    ID = 8

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='opponent_score',
            field_id=self.ID,
            base_type=BaseType.UINT16,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class EventFrontGearNumField(Field):
    ID = 9

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='front_gear_num',
            field_id=self.ID,
            base_type=BaseType.UINT8Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class EventFrontGearField(Field):
    ID = 10

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='front_gear',
            field_id=self.ID,
            base_type=BaseType.UINT8Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class EventRearGearNumField(Field):
    ID = 11

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='rear_gear_num',
            field_id=self.ID,
            base_type=BaseType.UINT8Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class EventRearGearField(Field):
    ID = 12

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='rear_gear',
            field_id=self.ID,
            base_type=BaseType.UINT8Z,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class EventDeviceIndexField(Field):
    ID = 13

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='device_index',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class EventRadarThreatLevelMaxField(Field):
    ID = 21

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='radar_threat_level_max',
            field_id=self.ID,
            base_type=BaseType.ENUM,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )


class EventRadarThreatCountField(Field):
    ID = 22

    def __init__(self, size: int = 0, growable: bool = True):
        super().__init__(
            name='radar_threat_count',
            field_id=self.ID,
            base_type=BaseType.UINT8,
            offset=0,
            scale=1,
            size=size,
            growable=growable,
            sub_fields=[
            ]
        )



import struct

from fit_tool.utils.crc import crc16


class ProtocolVersion:
    def __init__(self, major: int, minor: int):
        self.major = major
        self.minor = minor

    def to_bytes(self) -> bytes:
        return struct.pack('B', (self.major << 4) | self.minor)

    @classmethod
    def from_bytes(cls, bytes_buffer: bytes):
        value = bytes_buffer[0]
        major = value >> 4
        minor = value & 0x0f
        return ProtocolVersion(major, minor)

    def __str__(self):
        return f'{self.major}.{self.minor}'


class ProfileVersion:
    MAJOR_SCALE = 100

    def __init__(self, major: int, minor: int):
        self.major = major
        self.minor = minor

    def to_bytes(self) -> bytes:
        return struct.pack('<H', self.version_code)

    @property
    def version_code(self):
        return self.major * ProfileVersion.MAJOR_SCALE + self.minor

    @classmethod
    def from_bytes(cls, bytes_buffer: bytes):
        value, = struct.unpack('<H', bytes_buffer[0:2])
        major = value // ProfileVersion.MAJOR_SCALE
        minor = value % ProfileVersion.MAJOR_SCALE
        return ProfileVersion(major, minor)

    def __str__(self):
        return f'{self.major}.{self.minor}'


DEFAULT_PROTOCOL_VERSION = ProtocolVersion(2, 3)
DEFAULT_PROFILE_VERSION = ProfileVersion(21, 60)


class FitFileHeader:
    def __init__(self, records_size: int, protocol_version: ProtocolVersion = None,
                 profile_version: ProfileVersion = None, crc: int = None, gen_crc: bool = False):
        self.records_size = records_size
        self.protocol_version = protocol_version if protocol_version else DEFAULT_PROTOCOL_VERSION
        self.profile_version = profile_version if profile_version else DEFAULT_PROFILE_VERSION

        # crc16 of header bytes 0-11
        #
        # By including the CRC in the header you effectively reset the CRC for the
        # file, (when you CRC-16 a value with itself the CRC returned is 0)
        if crc is not None:
            self.crc = crc
        elif gen_crc:
            self.crc = FitFileHeader.generate_crc(self.protocol_version, self.profile_version, self.records_size)
        else:
            self.crc = None

    @classmethod
    def generate_crc(cls, protocol_version, profile_version, records_size):
        bytes_buffer = struct.pack('B', 14)
        bytes_buffer += protocol_version.to_bytes()
        bytes_buffer += profile_version.to_bytes()
        bytes_buffer += struct.pack('<I', records_size)
        bytes_buffer += b'.FIT'

        return crc16(bytes_buffer)

    @property
    def size(self) -> int:
        if self.crc is not None:
            return 14
        else:
            return 12

    def to_bytes(self) -> bytes:

        bytes_buffer = struct.pack('B', self.size)
        bytes_buffer += self.protocol_version.to_bytes()
        bytes_buffer += self.profile_version.to_bytes()
        bytes_buffer += struct.pack('<I', self.records_size)
        bytes_buffer += b'.FIT'

        if self.crc is not None:
            bytes_buffer += struct.pack('<H', self.crc)

        return bytes_buffer

    @classmethod
    def from_bytes(cls, bytes_buffer: bytes):
        offset = 0
        size, = struct.unpack('B', bytes_buffer[0:1])
        if size != len(bytes_buffer):
            raise Exception(f'Size {size} does not match bytes length: ${len(bytes_buffer)}')
        offset += 1

        protocol_version = ProtocolVersion.from_bytes(bytes_buffer[offset: offset + 1])
        offset += 1

        profile_version = ProfileVersion.from_bytes(bytes_buffer[offset: offset + 2])
        offset += 2

        records_size, = struct.unpack('<I', bytes_buffer[offset:offset + 4])
        offset += 4

        # .FIT
        tag_value, = struct.unpack('4s', bytes_buffer[offset:offset + 4])
        offset += 4
        if tag_value != b'.FIT':
            raise Exception('".FIT" not in header.')

        crc = None
        if len(bytes_buffer) == 14:
            crc, = struct.unpack('<H', bytes_buffer[offset:offset + 2])

        return cls(protocol_version=protocol_version,
                   profile_version=profile_version,
                   records_size=records_size,
                   crc=crc)
